#include "lpc2378.h"
#include "print.h"
#include "os.h"

// currentTaskId is a global variable that holds the ID of the current task.
//
// currentSP is a global variable that holds the value of the
// stack pointer for the current stack.

int currentTaskId;    // The ID of the current task
int currentSP;        // The value of the current task's stack pointer

// Array of TaskBlocks. One for each task.
TaskBlock taskBlocks[NUMBER_OF_TASKS];

/*
	Intialize stack.

	The stack grows down in memory so we want to start by calculating
	the top of the stack.

	Then fill in the initial data starting at the top and working down.

	The important values are the task_address and the initial CPSR.

	LR doesn't matter since the task is a infinite loop and should not exit.
	If it doesn't exit, then it doesn't use the LR since it doesn't return.

*/
int initialize_stack(int* stack, void* task_address)
{
	// Calculate the top of the stack.
	int *stkptr = (int*)(stack + STACKSIZE - 1);

	*(stkptr)   = (int)task_address;   // PC
	*(--stkptr) = (int)0x0e0e0e0e;     // R14 - LR
	*(--stkptr) = (int)0x0c0c0c0c;     // R12
	*(--stkptr) = (int)0x0b0b0b0b;     // R11
	*(--stkptr) = (int)0x0a0a0a0a;     // R10
	*(--stkptr) = (int)0x09090909;     // R9
	*(--stkptr) = (int)0x08080808;     // R8
	*(--stkptr) = (int)0x07070707;     // R7
	*(--stkptr) = (int)0x06060606;     // R6
	*(--stkptr) = (int)0x05050505;     // R5
	*(--stkptr) = (int)0x04040404;     // R4
	*(--stkptr) = (int)0x03030303;     // R3
	*(--stkptr) = (int)0x02020202;     // R2
	*(--stkptr) = (int)0x01010101;     // R1
	*(--stkptr) = (int)0x00000000;     // R0
	*(--stkptr) = ARM_SVC_MODE_ARM;    // CPSR

	return (int)(stkptr);
}

void CreateTask(int id, int* stack, void* task_address)
{
    taskBlocks[id].stackPointer = initialize_stack(stack, task_address);
    taskBlocks[id].delayCount = 0;
}

void StartOS(void)
{
    currentTaskId = 1;
    currentSP = taskBlocks[currentTaskId].stackPointer;

    // Run the current task
    switch_to_current();         // This call never returns to this function.
}

//
// Simple task scheduler that decides which task to run.
// It save the current stack pointer to the task block
// then picks a new task to run. If if cannot find a
// task to run it runs the idleTask.
//
void Scheduler(int sp)
{
	//>> Protect the body of this function with a critical section

	printString("Scheduler\n");

	// Save stack pointer for the current task.
    taskBlocks[currentTaskId].stackPointer = sp;

    int lastTaskId = currentTaskId;

    do
    {
    	currentTaskId++;
    	currentTaskId = currentTaskId < NUMBER_OF_TASKS - 1 ? currentTaskId : 0;
    } while(currentTaskId != lastTaskId && taskBlocks[currentTaskId].delayCount != 0);

    // If currentTaskId did not change then all the tasks are delayed
    // so run the idle task.
    if (currentTaskId == lastTaskId)
    {
    	// The idleTask's id is the last one.
    	currentTaskId = NUMBER_OF_TASKS - 1;
    }

    currentSP = taskBlocks[currentTaskId].stackPointer;
}

//
// TimerIrqHandler handles interrupts caused by the timer.
// It resets the timer then decrements the counter of any
// delay task. Then calls to schedule to switch tasks.
//
void TimerIrqHandler(int sp)
{
	// Reset timer
	WRITEREG32(T0IR, 0xFF);

	// Handle Ticks
    for(int i=0; i<NUMBER_OF_TASKS; i++)
    {
    	if(taskBlocks[i].delayCount > 0)
    	{
    		taskBlocks[i].delayCount--;
    	}
    }

	Scheduler(sp);
}

//
// DelayTask delays a task by setting it's delay counter
// then forcing a context switch.
//
void DelayTask(int taskId, int numOfTicks)
{
	//>> Protect with a critical section
    taskBlocks[taskId].delayCount = numOfTicks;
    SwitchContext();
}
